import { Graph } from './editor/d3/graph.ts';
import { GraphEditor } from './editor/GraphEditor.ts';

const IS_VALID = 'is-valid';
const IS_INVALID = 'is-invalid';

const wordInput = /** @type {HTMLInputElement} */ (document.getElementById('word-input'));
const buttons = /** @type {HTMLDivElement} */ (document.getElementById('input-buttons'));
const clearButton = /** @type {HTMLButtonElement} */ (document.getElementById('clear-button'));
const light = /** @type {HTMLDivElement} */ (document.getElementById('light'));

/**
 * @param {import('./examples.js').State[]} states
 */
export async function selectAutomaton(states) {
  let state = 0;
  let builder = '';

  const viewer = new GraphEditor();
  viewer.readonly = true;
  const graph = new Graph();
  for (let i = 0; i < states.length; i++) {
    const node = graph.createNode(i);
    if (states[i].accepting) {
      node.accepting = true;
    }
    if (i === 0) {
      node.start = true;
    }
  }
  for (let i = 0; i < states.length; i++) {
    const state = states[i];
    for (const [letter, target] of Object.entries(state.transitions)) {
      graph.createLink(i, target, letter);
    }
  }

  const container = /** @type {HTMLDivElement} */ (document.getElementById('state-graph'));
  container.appendChild(viewer);
  viewer.graph = graph;

  /**
   * Updates the UI to reflect the current state.
   */
  function updateUIState() {
    wordInput.value = builder;
    if (state === -1 || !states[state].accepting) {
      light.classList.remove(IS_VALID);
      light.classList.add(IS_INVALID);
    } else {
      light.classList.remove(IS_INVALID);
      light.classList.add(IS_VALID);
    }
    graph.forEach((node) => node.active = false);
    if (state in graph.nodes) {
      graph.nodes[state].active = true;
    }
    viewer.restart();
  }

  /**
   * Steps the FSM with the given letter.
   *
   * @param {string} letter
   */
  function step(letter) {
    if (state === -1) {
      return;
    }
    builder += letter;
    state = states[state].transitions[letter] ?? -1;
    updateUIState();
  }

  // Dynamically create buttons for each letter in the alphabet
  /**
   * @type {string[]}
   */
  const alphabet = Array.from(states.reduce((acc, current) => {
    Object.keys(current.transitions).forEach(current => acc.add(current));
    return acc;
  }, new Set())).sort();
  for (const letter of alphabet) {
    const button = document.createElement('button');
    button.innerText = letter;
    button.addEventListener('click', () => step(letter));
    buttons.appendChild(button);
  }

  // Reacts to input in the text box
  wordInput.addEventListener('input', () => {
    const value = wordInput.value;
    builder = '';
    state = 0;
    for (const letter of value) {
      step(letter);
    }
    if (!value.length) {
      updateUIState();
    }
  });

  clearButton.addEventListener('click', () => {
    wordInput.value = '';
    builder = '';
    state = 0;
    updateUIState();
  });

  updateUIState();
}
